/* 
 * File:   main.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 12. května 2019, 21:55
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <glib.h>

#include "gdgelement.h"
#include "gdgpart.h"
#include "gdgnet.h"
#include "vhdl.h"
#include "main.h"
#include "gdgpath.h"
#include "optimalizer.h"

#include "rsuk.h"


static gboolean main_connect_parts ( st_GDGPARTSALL *partall, st_GDGNETSALL *netall, gboolean not_fatal ) {

    int count_errors = 0;
    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];

        int j;
        for ( j = 0; j < net->count_pin; j++ ) {
            st_GDGNET_PIN *pin = net->pin[j];

            // pripojujeme jen parts (instance elementu), nikoliv PIN
            if ( pin->element ) {

                st_GDGPART *part = gdgparts_get_part_by_location ( partall, pin->loc_x, pin->loc_y, TRUE );

                if ( !part ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %d, pin: %d) can't found any part on this location (%s_%d_%d - B%04d) for net name: %s\n", __func__, __LINE__, i, j, pin->element->name, pin->loc_x, pin->loc_y, rsuk_get_rs_block_from_rs_loc ( rsuk_get_rs_loc_x ( pin->loc_x ), rsuk_get_rs_loc_y ( pin->loc_y ) ), net->name );
                    if ( !not_fatal ) return FALSE;
                    count_errors++;
                    continue;
                };

                if ( part->element != pin->element ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %d, pin: %d) found diferrent element (%s) on location %s_%d_%d (B%04d) for net name: %s\n", __func__, __LINE__, i, j, part->element->name, pin->element->name, pin->loc_x, pin->loc_y, rsuk_get_rs_block_from_rs_loc ( rsuk_get_rs_loc_x ( pin->loc_x ), rsuk_get_rs_loc_y ( pin->loc_y ) ), net->name );
                    if ( !not_fatal ) return FALSE;
                    count_errors++;
                    continue;
                };
                //printf ( "OK net: %d, pin: %d found element (%d) on location %s_%d_%d for net: %s\n", i, j, part->element->name, pin->element->name, pin->loc_x, pin->loc_y, net->name );

                st_GDGELEMENT_PORT *port = gdgelement_get_port_by_spice_netlist_pin_name ( pin->element, pin->part_pin_name );
                if ( !port ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %s, pin: %d, %s_%d_%d (B%04d) pin %s not found\n", __func__, __LINE__, net->name, j, pin->element->name, pin->loc_x, pin->loc_y, rsuk_get_rs_block_from_rs_loc ( rsuk_get_rs_loc_x ( pin->loc_x ), rsuk_get_rs_loc_y ( pin->loc_y ) ), pin->part_pin_name );
                    if ( !not_fatal ) return FALSE;
                    count_errors++;
                    continue;
                };

                if ( part->netpin[port->id] ) {
                    fprintf ( stderr, "%s(%d) - ERROR: (net: %s, pin: %d, %s_%d_%d (B%04d) pin %s is already connected to net name: %s\n", __func__, __LINE__, net->name, j, pin->element->name, pin->loc_x, pin->loc_y, rsuk_get_rs_block_from_rs_loc ( rsuk_get_rs_loc_x ( pin->loc_x ), rsuk_get_rs_loc_y ( pin->loc_y ) ), pin->part_pin_name, part->netpin[port->id]->name );
                    if ( !not_fatal ) return FALSE;
                    count_errors++;
                    continue;
                };

                // pripojime pin soucastky
                part->netpin[port->id] = net;
            };
        };
    };

    if ( count_errors ) fprintf ( stderr, "Part connector ending width %d error(s)\n\n", count_errors );

    return ( !count_errors ) ? TRUE : FALSE;
}


static gboolean main_error_on_part_connections ( st_GDGPARTSALL *partall, gboolean not_fatal ) {
    int count_errors = 0;
    int i;
    for ( i = 0; i < partall->count_part; i++ ) {
        st_GDGPART *part = partall->part[i];
        int j;
        for ( j = 0; j < part->element->count_port; j++ ) {
            if ( !part->netpin[j] ) {
                fprintf ( stderr, "%s(%d) - ERROR not connected port %s (%d) on part: %s_%d_%d (B%04d)\n", __func__, __LINE__, part->element->port[j]->spice_netlist_pin_name, j, part->element->name, part->loc_x, part->loc_y, rsuk_get_rs_block_from_rs_loc ( rsuk_get_rs_loc_x ( part->loc_x ), rsuk_get_rs_loc_y ( part->loc_y ) ) );
                if ( !not_fatal ) return TRUE;
                count_errors++;
            };
        };
    };
    if ( count_errors ) fprintf ( stderr, "Part connection tester ending width %d error(s)\n\n", count_errors );
    return ( count_errors ) ? TRUE : FALSE;
}


static gboolean main_error_on_net_connections ( st_GDGNETSALL *netall, gboolean not_fatal ) {
    int count_errors = 0;
    int i;
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];

        if ( !net->is_constant ) {
            st_GDGNET_PIN *pin = gdgnet_get_net_source_pin ( net );

            if ( !pin ) {
                fprintf ( stderr, "%s(%d) - ERROR net without power source %s\n", __func__, __LINE__, net->name );
                if ( !not_fatal ) return TRUE;
                count_errors++;
            };
        };
    };
    if ( count_errors ) fprintf ( stderr, "Net connection tester ending width %d error(s)\n\n", count_errors );
    return ( count_errors ) ? TRUE : FALSE;
}

#define DEFAULT_VHDL_MODULE_NAME    "GDG_internal"

static char *spice_filepath = "testGDG_mod.spice";
static char *netlist_filepath = "testGDG_mod_exp.txt";
static char *output_filepath = NULL;
static char *path_print_to_dest = NULL;
//static char *path_vhdl_to_dest = "CPU";
static char *path_vhdl_to_dest = NULL;
static char *path_elements_to_dest = NULL;
static gboolean create_full_vhdl = TRUE;
static gboolean path_verbose = FALSE;
static gboolean optimized_output = FALSE;
static gboolean no_fatal = FALSE;
static char *vhdl_module_name = DEFAULT_VHDL_MODULE_NAME;

static char *rsuk_netlist_filepath = NULL;
static int rsuk_compare_method = 0;

/*
static char *rsuk_netlist_filepath = "cesty_blok.txt";
static gboolean rsuk_compare = TRUE;
 */

static GOptionEntry entries[] = {
    { "spice", 's', 0, G_OPTION_ARG_FILENAME, &spice_filepath, "source SPICE file (testGDG_mod.spice)", "filepath" },
    { "nets", 'n', 0, G_OPTION_ARG_FILENAME, &netlist_filepath, "source Netlist file (testGDG_mod_exp.txt)", "filepath" },
    { "output", 'o', 0, G_OPTION_ARG_FILENAME, &output_filepath, "output file - using stdout if is not set", "output_file" },
    { "vhdl", 'l', 0, G_OPTION_ARG_NONE, &create_full_vhdl, "create VHDL entity GDG_internal (default)", NULL },
    { "path", 'p', 0, G_OPTION_ARG_STRING, &path_print_to_dest, "show path to destination net", "net_name" },
    { "path-elements", 'e', 0, G_OPTION_ARG_STRING, &path_elements_to_dest, "show elements of path to destination net", "net_name" },
    { "vhdl-path", 'd', 0, G_OPTION_ARG_STRING, &path_vhdl_to_dest, "create VHDL entity for destination net", "net_name" },
    { "path-verbose", 'v', 0, G_OPTION_ARG_NONE, &path_verbose, "path verbose mode", NULL },
    { "optimized", 'z', 0, G_OPTION_ARG_NONE, &optimized_output, "create optimized output (no buffers and use logic primitives)", NULL },
    { "module-name", 'm', 0, G_OPTION_ARG_STRING, &vhdl_module_name, "VHDL module name", DEFAULT_VHDL_MODULE_NAME },

    { "rsuk", 'r', 0, G_OPTION_ARG_FILENAME, &rsuk_netlist_filepath, "load RSUK netlist", "filepath" },
    { "compare-method", 'c', 0, G_OPTION_ARG_INT, &rsuk_compare_method, "compare RSUK and SPICE netlist", "<1,2>" },
    { "nonstop", 'i', 0, G_OPTION_ARG_NONE, &no_fatal, "non stop on connection errors", NULL },
    { NULL }
};

GOptionContext *context;


static void main_parse_options ( int argc, char** argv ) {

    GError *error = NULL;

    context = g_option_context_new ( "- convertor of GDG spice files to VHDL" );

    g_option_context_add_main_entries ( context, entries, NULL );
    if ( !g_option_context_parse ( context, &argc, &argv, &error ) ) {
        g_print ( "option parsing failed: %s\n", error->message );
        exit ( 1 );
    };

    if ( path_print_to_dest || path_vhdl_to_dest || path_elements_to_dest ) create_full_vhdl = FALSE;
}


void main_error_exit ( char *error_message ) {
    fprintf ( stderr, "\n%s\n\n", error_message );
    fprintf ( stderr, "%s", g_option_context_get_help ( context, TRUE, NULL ) );
    exit ( EXIT_FAILURE );
}


static void main_print_version ( void ) {
    GString *str = g_string_new ( "" );
    g_string_append_printf ( str, "%s, %s\n", MAIN_PROGRAM_NAME, MAIN_PROGRAM_VERSION );
    fprintf ( stderr, "%s", str->str );
    g_string_free ( str, TRUE );
}


/*
 * 
 */
int main ( int argc, char** argv ) {

    main_print_version ( );

    main_parse_options ( argc, argv );

    st_GDGELEMENTSALL *elmall = NULL;
    st_GDGPARTSALL *partall = NULL;
    st_GDGNETSALL *netall = NULL;

    st_GDGELEMENTSALL *spice_elmall = gdgelements_init ( TRUE );
    if ( !spice_elmall ) return EXIT_FAILURE;

    st_GDGPARTSALL *spice_partall = gdgparts_init ( spice_elmall, spice_filepath );
    if ( !spice_partall ) return EXIT_FAILURE;

    st_GDGNETSALL *spice_netall = gdgnets_init ( spice_elmall, netlist_filepath );
    if ( !spice_netall ) return EXIT_FAILURE;

    gboolean err = main_connect_parts ( spice_partall, spice_netall, no_fatal ) ? FALSE : TRUE;
    if ( ( no_fatal ) || ( !err ) ) err |= main_error_on_part_connections ( spice_partall, no_fatal );
    if ( ( no_fatal ) || ( !err ) ) err |= main_error_on_net_connections ( spice_netall, no_fatal );

    st_GDGELEMENTSALL *rsuk_elmall = NULL;
    st_GDGPARTSALL *rsuk_partall = NULL;
    st_GDGNETSALL *rsuk_netall = NULL;

    if ( ( no_fatal ) || ( !err ) ) {
        if ( rsuk_netlist_filepath ) {

            rsuk_elmall = gdgelements_init ( FALSE );
            if ( !rsuk_elmall ) return EXIT_FAILURE;

            rsuk_partall = rsuk_gdgparts_init ( rsuk_elmall, rsuk_netlist_filepath );
            if ( !rsuk_partall ) return EXIT_FAILURE;

            rsuk_netall = rsuk_gdgnets_init ( rsuk_elmall, rsuk_netlist_filepath );
            if ( !rsuk_netall ) return EXIT_FAILURE;

            err |= main_connect_parts ( rsuk_partall, rsuk_netall, no_fatal ) ? FALSE : TRUE;
            if ( ( no_fatal ) || ( !err ) ) err |= main_error_on_part_connections ( rsuk_partall, no_fatal );
            if ( ( no_fatal ) || ( !err ) ) err |= main_error_on_net_connections ( rsuk_netall, no_fatal );
        };
    };

    if ( !err ) {

        if ( rsuk_netall ) {
            fprintf ( stderr, "Using RSUK data!\n" );
            elmall = rsuk_elmall;
            partall = rsuk_partall;
            netall = rsuk_netall;
        } else {
            fprintf ( stderr, "Using SPICE data!\n" );
            elmall = spice_elmall;
            partall = spice_partall;
            netall = spice_netall;
        };

        if ( rsuk_compare_method ) {
            if ( !rsuk_netall ) {
                fprintf ( stderr, "ERROR - RSUK netlist missing!\n" );
            } else {
                if ( rsuk_compare_method == 1 ) {
                    rsuk_compare_parts ( spice_partall, rsuk_partall );
                    rsuk_compare_netlist ( spice_partall, rsuk_partall );
                } else if ( rsuk_compare_method == 2 ) {
                    rsuk_compare_parts ( spice_partall, rsuk_partall );
                    rsuk_compare_netlist ( rsuk_partall, spice_partall );
                } else {
                    fprintf ( stderr, "ERROR - invalid compare method '%d'!\n", rsuk_compare_method );
                };
            };

        } else {

            if ( optimized_output ) {
                optimalizer_run ( elmall, partall, netall );
            };

            st_GDGIOPORTSALL *ioportall = gdgnets_create_ioports ( netall, NULL );

            if ( create_full_vhdl ) {
                /* full GDG vhdl output */
                char *vhdl_txt = vhdl_create ( vhdl_module_name, partall, netall, ioportall, optimized_output );
                if ( !output_filepath ) {
                    printf ( "%s", vhdl_txt );
                } else {
                    FILE *fp = fopen ( output_filepath, "w" );
                    if ( !fp ) {
                        GString *str = g_string_new ( "" );
                        g_string_append_printf ( str, "%s(%d) - ERROR: Can't open output file '%s'\n", __func__, __LINE__, output_filepath );
                        main_error_exit ( g_string_free ( str, FALSE ) );
                        return EXIT_FAILURE;
                    };

                    fprintf ( fp, "%s", vhdl_txt );
                    fclose ( fp );
                };
                g_free ( vhdl_txt );

            } else {

                char *search_net_name = "";

                if ( path_print_to_dest ) {
                    search_net_name = path_print_to_dest;
                } else if ( path_elements_to_dest ) {
                    search_net_name = path_elements_to_dest;
                } else if ( path_vhdl_to_dest ) {
                    search_net_name = path_vhdl_to_dest;
                };

                st_GDGNET *net = gdgnets_get_net_by_name ( netall, search_net_name );

                if ( !net ) {
                    GString *str = g_string_new ( "" );
                    g_string_append_printf ( str, "%s(%d) - ERROR: can't found this net name '%s'\n", __func__, __LINE__, search_net_name );
                    main_error_exit ( g_string_free ( str, FALSE ) );
                    return EXIT_FAILURE;
                };

                fprintf ( stderr, "OK, net name '%s' found!\n", search_net_name );

                gboolean create_path_str = ( path_print_to_dest ) ? TRUE : FALSE;

                st_GDGPATH *path = gdgpath_create_for_net ( partall, net, create_path_str, path_verbose );
                if ( !path ) {
                    fprintf ( stderr, "ERROR, path not found!\n" );
                } else {

                    GString *str = g_string_new ( "DONE, complete path found!\n" );
                    if ( path->create_path_str ) {
                        g_string_append ( path->str, str->str );
                    };
                    if ( path->verbose ) {
                        fprintf ( stderr, "%s", str->str );
                    };
                    g_string_free ( str, TRUE );

                    char *txt = NULL;

                    if ( path_print_to_dest ) {
                        GString *str = g_string_new ( "" );
                        g_string_append_printf ( str, "\nPath to net name '%s':\n%s", path->dst_net->name, path->str->str );
                        g_string_append ( str, path->str->str );
                        txt = g_string_free ( str, FALSE );
                    } else if ( path_elements_to_dest ) {
                        txt = gdgpath_get_elements_txt ( path );
                    } else if ( path_vhdl_to_dest ) {
                        txt = gdgpath_get_vhdl ( path, optimized_output );
                    };

                    if ( !output_filepath ) {
                        printf ( "%s", txt );
                    } else {
                        FILE *fp = fopen ( output_filepath, "w" );
                        if ( !fp ) {
                            GString *str = g_string_new ( "" );
                            g_string_append_printf ( str, "%s(%d) - ERROR: Can't open output file '%s'\n", __func__, __LINE__, output_filepath );
                            main_error_exit ( g_string_free ( str, FALSE ) );
                            return EXIT_FAILURE;
                        };

                        fprintf ( fp, "%s", txt );
                        fclose ( fp );
                    };
                    g_free ( txt );

                    gdgpath_destroy ( path );
                };
            };

            gdgnets_ioports_destroy ( ioportall );
        };
    };


    gdgparts_destroy ( spice_partall );
    gdgnets_destroy ( spice_netall );
    gdgelements_destroy ( spice_elmall );

    gdgparts_destroy ( rsuk_partall );
    gdgnets_destroy ( rsuk_netall );
    gdgelements_destroy ( rsuk_elmall );

    return EXIT_SUCCESS;
}

